@@ -2,6 +2,7 @@ from typing import Callable, Dict, Tuple
 
 import aesara
 import aesara.tensor as at
+import numpy as np
 from aesara.ifelse import ifelse
 from aesara.scan.utils import until
 from aesara.tensor.random.utils import RandomStream
@@ -34,11 +35,54 @@ def static_integration(
     integrator: Callable,
     num_integration_steps: int,
 ) -> Callable:
-    """Generate a trajectory by integrating several times in one direction."""
+    """Build a function that generates fixed-length trajectories.
+
+    Parameters
+    ----------
+    integrator
+       Function that performs one integration step.
+    num_integration_steps
+        The number of times we need to run the integrator every time the
+        returned function is called.
+
+    Returns
+    -------
+    A function that integrates the hamiltonian dynamics a
+    `num_integration_steps` times.
+
+    """
 
     def integrate(
-        q_init, p_init, energy_init, energy_grad_init, step_size
+        q_init: TensorVariable,
+        p_init: TensorVariable,
+        energy_init: TensorVariable,
+        energy_grad_init: TensorVariable,
+        step_size: TensorVariable,
     ) -> Tuple[IntegratorStateType, Dict]:
+        """Generate a trajectory by integrating several times in one direction.
+
+        Parameters
+        ----------
+        q_init
+            The initial position.
+        p_init
+            The initial value of the momentum.
+        energy_init
+            The initial value of the potential energy.
+        energy_grad_init
+            The initial value of the gradient of the potential energy wrt the position.
+        step_size
+            The size of each step taken by the integrator.
+
+        Returns
+        -------
+        A tuple with the last position, value of the momentum, potential energy,
+        gradient of the potential energy wrt the position in a tuple as well as
+        a dictionary that contains the update rules for all the shared variables
+        updated in `scan`.
+
+        """
+
         def one_step(q, p, potential_energy, potential_energy_grad):
             new_state = integrator(
                 q, p, potential_energy, potential_energy_grad, step_size
@@ -83,8 +127,10 @@ def dynamic_integration(
 
     Parameters
     ----------
+    srng
+        A RandomStream object that tracks the changes in a shared random state.
     integrator
-        The symplectic integrator used to integrate the hamiltonian trajectory.
+        The symplectic integrator used to integrate the hamiltonian dynamics.
     kinetic_energy
         Function to compute the current value of the kinetic energy.
     update_termination_state
@@ -94,6 +140,11 @@ def dynamic_integration(
     divergence_threshold
         Value of the difference of energy between two consecutive states above which we say a transition is divergent.
 
+    Returns
+    -------
+    A function that integrates the trajectory in one direction and updates a
+    proposal until the termination criterion is met.
+
     """
     generate_proposal = proposal_generator(kinetic_energy, divergence_threshold)
     sample_proposal = progressive_uniform_sampling
@@ -111,8 +162,6 @@ def dynamic_integration(
 
         Parameters
         ----------
-        rng_key
-            Key used by JAX's random number generator.
         previous_last_state
             The last state of the previously integrated trajectory.
         direction int in {-1, 1}
@@ -128,6 +177,17 @@ def dynamic_integration(
         initial_energy
             Initial energy H0 of the HMC step (not to confused with the initial energy of the subtree)
 
+        Returns
+        -------
+        A tuple with on the one hand: a new proposal (sampled from the states
+        traversed while building the trajectory), the last state, the sum of the
+        momenta values along the trajectory (needed for termination criterion),
+        the updated termination state, the number of integration steps
+        performed, a boolean that indicates whether the trajectory has diverged,
+        a boolean that indicates whether the termination criterion was met. And
+        on the other hand a dictionary that contains the update rules for the
+        shared variables updated in the internal `Scan` operator.
+
         """
 
         def add_one_state(
@@ -206,9 +266,21 @@ def dynamic_integration(
             state[1],
             0,
         )
+        full_initial_state = (
+            *proposal[0],
+            proposal[1],
+            proposal[2],
+            proposal[3],
+            *state,
+            momentum_sum,
+            *termination_state,
+            at.as_tensor(1, dtype=np.int32),
+            is_diverging,
+            np.array(False),
+        )
 
         steps = at.arange(1, 1 + max_num_steps)
-        traj, updates = aesara.scan(
+        trajectory, updates = aesara.scan(
             add_one_state,
             outputs_info=(
                 *proposal[0],
@@ -218,25 +290,34 @@ def dynamic_integration(
                 *state,
                 momentum_sum,
                 *termination_state,
-                at.as_tensor(1, dtype="int32"),
+                at.as_tensor(1, dtype=np.int32),
                 None,
                 None,
             ),
             sequences=steps,
         )
+        full_last_state = tuple([state[-1] for state in trajectory])
+
+        # We build the trajectory iff the first step is not diverging
+        full_state = ifelse(is_diverging, full_initial_state, full_last_state)
 
         new_proposal = (
-            (traj[0][-1], traj[1][-1], traj[2][-1], traj[3][-1]),
-            traj[4][-1],
-            traj[5][-1],
-            traj[6][-1],
+            (full_state[0], full_state[1], full_state[2], full_state[3]),
+            full_state[4],
+            full_state[5],
+            full_state[6],
         )
-        new_state = (traj[7][-1], traj[8][-1], traj[9][-1], traj[10][-1])
-        subtree_momentum_sum = traj[11][-1]
-        new_termination_state = (traj[12][-1], traj[13][-1], traj[14][-1], traj[15][-1])
-        trajectory_length = traj[-3][-1]  # defined as the number of steps taken
-        is_diverging = traj[-2][-1] | is_diverging
-        has_terminated = traj[-1][-1]
+        new_state = (full_state[7], full_state[8], full_state[9], full_state[10])
+        subtree_momentum_sum = full_state[11]
+        new_termination_state = (
+            full_state[12],
+            full_state[13],
+            full_state[14],
+            full_state[15],
+        )
+        trajectory_length = full_state[-3]
+        is_diverging = full_state[-2]
+        has_terminated = full_state[-1]
 
         return (
             new_proposal,
@@ -269,6 +350,8 @@ def multiplicative_expansion(
 
     Parameters
     ----------
+    srng
+        A RandomStream object that tracks the changes in a shared random state.
     trajectory_integrator
         A function that runs the symplectic integrators and returns a new proposal
         and the integrated trajectory.
@@ -290,6 +373,32 @@ def multiplicative_expansion(
         initial_energy,
         step_size,
     ):
+        """Expand the current trajectory multiplicatively.
+
+        At each step we draw a direction at random, build a subtrajectory starting
+        from the leftmost or rightmost point of the current trajectory that is
+        twice as long as the current trajectory.
+
+        Once that is done, possibly update the current proposal with that of
+        the subtrajectory.
+
+        Parameters
+        ----------
+        proposal
+            Current new state proposal.
+        left_state
+            The current leftmost state of the trajectory.
+        right_state
+            The current rightmost state of the trajectory.
+        momentum_sum
+            The current value of the sum of momenta along the trajectory.
+        initial_energy
+            Potential energy before starting to build the trajectory.
+        step_size
+            The size of each step taken by the integrator.
+
+        """
+
         def expand_once(
             step,
             q_proposal,  # proposal
@@ -313,16 +422,6 @@ def multiplicative_expansion(
             idx_min,
             idx_max,
         ):
-            """Expand the current trajectory.
-
-            At each step we draw a direction at random, build a subtrajectory starting
-            from the leftmost or rightmost point of the current trajectory that is
-            twice as long as the current trajectory.
-
-            Once that is done, possibly update the current proposal with that of
-            the subtrajectory.
-
-            """
 
             left_state = (
                 q_left,
@@ -356,7 +455,7 @@ def multiplicative_expansion(
 
             do_go_right = srng.bernoulli(0.5)
             direction = at.where(do_go_right, 1.0, -1.0)
-            start_state = where_state(do_go_right, right_state, left_state)
+            start_state = ifelse(do_go_right, right_state, left_state)
 
             (
                 new_proposal,
@@ -378,8 +477,8 @@ def multiplicative_expansion(
             # Update the trajectory.
             # The trajectory integrator always integrates forward in time; we
             # thus need to switch the states if the other direction was picked.
-            new_left_state = where_state(do_go_right, left_state, new_state)
-            new_right_state = where_state(do_go_right, new_state, right_state)
+            new_left_state = ifelse(do_go_right, left_state, new_state)
+            new_right_state = ifelse(do_go_right, new_state, right_state)
             new_momentum_sum = momentum_sum + subtree_momentum_sum
 
             # Compute the pseudo-acceptance probability for the NUTS algorithm.
@@ -456,27 +555,6 @@ def multiplicative_expansion(
     return expand
 
 
-def where_state(
-    do_pick_left: bool,
-    left_state: IntegratorStateType,
-    right_state: IntegratorStateType,
-) -> IntegratorStateType:
-    """Represents a switch between two states depending on a condition."""
-    q_left, p_left, potential_energy_left, potential_energy_grad_left = left_state
-    q_right, p_right, potential_energy_right, potential_energy_grad_right = right_state
-
-    q = ifelse(do_pick_left, q_left, q_right)
-    p = ifelse(do_pick_left, p_left, p_right)
-    potential_energy = at.where(
-        do_pick_left, potential_energy_left, potential_energy_right
-    )
-    potential_energy_grad = ifelse(
-        do_pick_left, potential_energy_grad_left, potential_energy_grad_right
-    )
-
-    return (q, p, potential_energy, potential_energy_grad)
-
-
 def where_proposal(
     do_pick_left: bool,
     left_proposal: ProposalStateType,
@@ -486,7 +564,7 @@ def where_proposal(
     left_state, left_weight, left_energy, left_log_sum_p_accept = left_proposal
     right_state, right_weight, right_energy, right_log_sum_p_accept = right_proposal
 
-    state = where_state(do_pick_left, left_state, right_state)
+    state = ifelse(do_pick_left, left_state, right_state)
     energy = at.where(do_pick_left, left_energy, right_energy)
     weight = at.where(do_pick_left, left_weight, right_weight)
     log_sum_p_accept = ifelse(
